"""Managers for character section models (1/3)."""
# pylint: disable=missing-class-docstring

from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, Set

from django.db import DatabaseError, models, transaction
from django.db.models import Case, ExpressionWrapper, F, Value, When
from esi.models import Token
from eveuniverse.models import EveEntity, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
from memberaudit.decorators import fetch_token_for_character
from memberaudit.helpers import (
    UpdateSectionResult,
    data_retention_cutoff,
    eve_entity_ids_from_objs,
    model_to_dict_safely,
)
from memberaudit.models._helpers import store_character_data_to_disk_when_enabled
from memberaudit.providers import esi
from memberaudit.utils import (
    get_or_create_esi_or_none,
    get_or_create_or_none,
    get_or_none,
)

from ._common import GenericUpdateSimpleObjMixin

if TYPE_CHECKING:
    from memberaudit.models import Character, CharacterAsset


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterAssetQuerySet(models.QuerySet):
    def item_ids(self) -> Set[int]:
        """Return item IDs for objects in this queryset."""
        return set(self.values_list("item_id", flat=True))

    def annotate_pricing(self) -> models.QuerySet:
        """Return qs with annotated price and total columns."""
        return (
            self.select_related("eve_type__market_price")
            .annotate(
                price=Case(
                    When(
                        is_blueprint_copy=True,
                        then=Value(None),
                    ),
                    default=F("eve_type__market_price__average_price"),
                )
            )
            .annotate(
                total=Case(
                    When(
                        is_blueprint_copy=True,
                        then=Value(None),
                    ),
                    default=ExpressionWrapper(
                        F("eve_type__market_price__average_price") * F("quantity"),
                        output_field=models.FloatField(),
                    ),
                )
            )
        )


class CharacterAssetManagerBase(models.Manager):
    def fetch_from_esi(self, character: Character, force_update: bool = False):
        """Fetch assets from ESI and preload related objects from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.ASSETS,
            fetch_func=self._fetch_data_from_esi,
            store_func=None,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-assets.read_assets.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> list:
        """Fetch character assets with names from ESI and return it."""
        asset_data = self._fetching_assets_from_esi(character, token)

        # add names to assets
        item_ids = list(asset_data.keys())
        asset_names = self._fetching_asset_names_from_esi(character, token, item_ids)
        for item_id in item_ids:
            asset_data[item_id]["name"] = asset_names.get(item_id, "")

        return sorted(asset_data.values(), key=lambda o: o["item_id"])

    def _fetching_assets_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching assets from ESI", character)
        asset_list = esi.client.Assets.get_characters_character_id_assets(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        asset_data = {int(item["item_id"]): item for item in asset_list}
        return asset_data

    def _fetching_asset_names_from_esi(
        self, character: Character, token: Token, item_ids: List[int]
    ):
        logger.info("%s: Fetching asset names from ESI", character)
        names = []
        for asset_ids_chunk in chunks(item_ids, 999):
            names += esi.client.Assets.post_characters_character_id_assets_names(
                character_id=character.eve_character.character_id,
                token=token.valid_access_token(),
                item_ids=asset_ids_chunk,
            ).results()

        asset_names = {
            int(item["item_id"]): item["name"]
            for item in names
            if item["name"] != "None"
        }

        return asset_names

    def preload_objects_from_esi(
        self, character: Character, asset_list: list
    ) -> UpdateSectionResult:
        """Preloads objects needed to build the asset tree."""

        logger.info("%s: Preloading objects for asset tree", character)

        if not asset_list:
            return UpdateSectionResult(is_changed=None, is_updated=False)

        is_updated = self._fetch_missing_types_from_esi(character, asset_list)
        is_updated |= self._create_missing_locations(character, asset_list)

        return UpdateSectionResult(is_changed=None, is_updated=is_updated)

    def _fetch_missing_types_from_esi(
        self, character: Character, asset_list: list
    ) -> bool:
        required_ids = {item["type_id"] for item in asset_list if "type_id" in item}
        existing_ids = set(EveType.objects.values_list("id", flat=True))
        missing_ids = required_ids.difference(existing_ids)
        if not missing_ids:
            return False

        EveType.objects.bulk_get_or_create_esi(ids=list(missing_ids))
        logger.info(
            "%s: Fetched %s missing types from ESI", character, len(missing_ids)
        )
        return True

    def _create_missing_locations(self, character: Character, asset_list: list) -> bool:
        from memberaudit.models import Location

        asset_item_ids = {asset["item_id"] for asset in asset_list}
        incoming_location_ids = {
            item["location_id"]
            for item in asset_list
            if "location_id" in item and item["location_id"] not in asset_item_ids
        }
        if not incoming_location_ids:
            return False

        current_location_ids = set(Location.objects.values_list("id", flat=True))
        missing_location_ids = incoming_location_ids - current_location_ids
        if not missing_location_ids:
            return False

        token = character.fetch_token("esi-universe.read_structures.v1")
        Location.objects.create_missing_esi(
            location_ids=missing_location_ids, token=token
        )
        return True

    def bulk_create_with_fallback(
        self, objs: Iterable[CharacterAsset], batch_size: int = None
    ) -> List[CharacterAsset]:
        """Create objs in bulk safely and return newly created objs."""
        try:
            added_objs = self.bulk_create(objs, batch_size=batch_size)
        except DatabaseError:
            logger.warning(
                "Bulk create with %d %s failed. "
                "Falling back on creating them one by one.",
                len(objs),
                self.model._meta.verbose_name_plural,
                exc_info=True,
            )
            added_objs = []
            for obj in objs:
                try:
                    obj.save(force_insert=True)
                except DatabaseError:
                    obj_as_dict = model_to_dict_safely(obj)
                    logger.exception(
                        "Failed to create %s: %s",
                        self.model._meta.verbose_name,
                        obj_as_dict,
                    )
                else:
                    added_objs.append(obj)

        return added_objs


CharacterAssetManager = CharacterAssetManagerBase.from_queryset(CharacterAssetQuerySet)


class CharacterAttributesManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create attributes for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.ATTRIBUTES,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-skills.read_skills.v1")
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching attributes from ESI", character)
        attribute_data = esi.client.Skills.get_characters_character_id_attributes(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return attribute_data

    def _update_or_create_objs(self, character: Character, attribute_data):
        self.update_or_create(
            character=character,
            defaults={
                "accrued_remap_cooldown_date": attribute_data.get(
                    "accrued_remap_cooldown_date"
                ),
                "last_remap_date": attribute_data.get("last_remap_date"),
                "bonus_remaps": attribute_data.get("bonus_remaps"),
                "charisma": attribute_data.get("charisma"),
                "intelligence": attribute_data.get("intelligence"),
                "memory": attribute_data.get("memory"),
                "perception": attribute_data.get("perception"),
                "willpower": attribute_data.get("willpower"),
            },
        )


class CharacterContactLabelManager(GenericUpdateSimpleObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create assets for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.CONTACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
            hash_num=2,
        )

    @fetch_token_for_character("esi-characters.read_contacts.v1")
    def _fetch_data_from_esi(self, character: Character, token) -> List[dict]:
        logger.info("%s: Fetching contact labels from ESI", character)
        labels = esi.client.Contacts.get_characters_character_id_contacts_labels(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return labels

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character, key, value):
            obj = self.model(character=character, label_id=key, name=value)
            return obj

        self._update_or_create_objs_generic(
            character,
            esi_data,
            esi_fields=("label_id", "label_name"),
            model_fields=("label_id", "name"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
        )


class CharacterContactManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create assets for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.CONTACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_contacts.v1")
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching contacts from ESI", character)
        contacts_data = esi.client.Contacts.get_characters_character_id_contacts(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return contacts_data

    @transaction.atomic()
    def _update_or_create_objs(self, character: Character, contacts_data) -> Set[int]:
        """Update or create new contact objects from provided data.

        Return EveEntity IDs in newly created contacts.
        """
        contacts_list = (
            {int(obj["contact_id"]): obj for obj in contacts_data}
            if contacts_data
            else {}
        )
        incoming_ids = set(contacts_list.keys())
        existing_ids = set(
            self.filter(character=character).values_list("eve_entity_id", flat=True)
        )
        obsolete_ids = existing_ids.difference(incoming_ids)
        if obsolete_ids:
            logger.info(
                "%s: Removing %s obsolete contacts", character, len(obsolete_ids)
            )
            self.filter(character=character, eve_entity_id__in=obsolete_ids).delete()

        create_ids = incoming_ids.difference(existing_ids)
        if create_ids:
            self._create_new_contacts(
                character=character,
                contacts_list=contacts_list,
                contact_ids=list(create_ids),
            )

        update_ids = incoming_ids.difference(create_ids)
        if update_ids:
            self._update_existing_contacts(
                character=character,
                contacts_list=contacts_list,
                contact_ids=list(update_ids),
            )

        if not obsolete_ids and not create_ids and not update_ids:
            logger.info("%s: Contacts have not changed", character)

        return create_ids

    def _create_new_contacts(
        self, character: Character, contacts_list: dict, contact_ids: list
    ):
        logger.info("%s: Storing %s new contacts", character, len(contact_ids))
        new_contacts_list = {
            contact_id: obj
            for contact_id, obj in contacts_list.items()
            if contact_id in contact_ids
        }
        new_contacts = [
            self.model(
                character=character,
                eve_entity=get_or_create_or_none("contact_id", contact_data, EveEntity),
                is_blocked=contact_data.get("is_blocked"),
                is_watched=contact_data.get("is_watched"),
                standing=contact_data.get("standing"),
            )
            for contact_data in new_contacts_list.values()
        ]
        self.bulk_create(new_contacts, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        self._update_contact_contact_labels(
            character=character,
            contacts_list=contacts_list,
            contact_ids=contact_ids,
            is_new=True,
        )

    def _update_contact_contact_labels(
        self, character: Character, contacts_list: dict, contact_ids: list, is_new=False
    ):
        from memberaudit.models import CharacterContactLabel

        for contact_id, contact_data in contacts_list.items():
            if contact_id in contact_ids and contact_data.get("label_ids"):
                character_contact = self.filter(character=character).get(
                    eve_entity_id=contact_id
                )
                if not is_new:
                    character_contact.labels.clear()

                labels = []
                for label_id in contact_data.get("label_ids"):
                    try:
                        label = character.contact_labels.get(label_id=label_id)
                    except CharacterContactLabel.DoesNotExist:
                        # sometimes label IDs on contacts
                        # do not refer to actual labels
                        logger.info(
                            "%s: Unknown contact label with id %s",
                            character,
                            label_id,
                        )
                    else:
                        labels.append(label)

                    character_contact.labels.add(*labels)

    def _update_existing_contacts(
        self, character: Character, contacts_list: dict, contact_ids: list
    ):
        logger.info("%s: Updating %s contacts", character, len(contact_ids))
        update_contact_pks = list(
            self.filter(character=character, eve_entity_id__in=contact_ids).values_list(
                "pk", flat=True
            )
        )
        contacts = self.in_bulk(update_contact_pks)
        for contact in contacts.values():
            contact_data = contacts_list.get(contact.eve_entity_id)
            if contact_data:
                contact.is_blocked = contact_data.get("is_blocked")
                contact.is_watched = contact_data.get("is_watched")
                contact.standing = contact_data.get("standing")

        self.bulk_update(
            contacts.values(),
            fields=["is_blocked", "is_watched", "standing"],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )
        self._update_contact_contact_labels(
            character=character, contacts_list=contacts_list, contact_ids=contact_ids
        )


class CharacterContractManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create contracts for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.CONTRACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        logger.info("%s: Fetching contracts from ESI", character)
        contracts_data = esi.client.Contracts.get_characters_character_id_contracts(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        store_character_data_to_disk_when_enabled(
            character=character, data=contracts_data, section="contracts"
        )

        cutoff_datetime = data_retention_cutoff()
        contracts_list = {
            obj.get("contract_id"): obj
            for obj in contracts_data
            if cutoff_datetime is None or obj.get("date_expired") > cutoff_datetime
        }
        return contracts_list

    @fetch_token_for_character("esi-universe.read_structures.v1")
    def _update_or_create_objs(
        self, character: Character, token: Token, contracts_list
    ) -> Set[int]:
        """Update or create new contract objects from provided data.

        Return EveEntity IDs in newly created contracts.
        """
        from memberaudit.models import Location

        if cutoff_datetime := data_retention_cutoff():
            character.contracts.filter(date_expired__lt=cutoff_datetime).delete()

        existing_ids = set(character.contracts.values_list("contract_id", flat=True))
        incoming_location_ids = {
            obj["start_location_id"]
            for contract_id, obj in contracts_list.items()
            if contract_id not in existing_ids
        }
        incoming_location_ids |= {
            obj["end_location_id"] for obj in contracts_list.values()
        }
        Location.objects.create_missing_esi(incoming_location_ids, token)

        with transaction.atomic():
            incoming_ids = set(contracts_list.keys())
            existing_ids = set(
                self.filter(character=character).values_list("contract_id", flat=True)
            )
            create_ids = incoming_ids.difference(existing_ids)
            if create_ids:
                new_entity_ids = self._create_new_contracts(
                    character=character,
                    contracts_list=contracts_list,
                    contract_ids=create_ids,
                )
            else:
                new_entity_ids = set()

            update_ids = incoming_ids.difference(create_ids)
            if update_ids:
                self._update_existing_contracts(
                    character=character,
                    contracts_list=contracts_list,
                    contract_ids=update_ids,
                )
            return new_entity_ids

    def _create_new_contracts(
        self, character: Character, contracts_list: dict, contract_ids: Set[int]
    ) -> Set[int]:
        from memberaudit.models import Location

        logger.info("%s: Storing %s new contracts", character, len(contract_ids))
        new_contracts = []
        for contract_id in contract_ids:
            contract_data = contracts_list.get(contract_id)
            if contract_data:
                new_contracts.append(
                    self.model(
                        character=character,
                        contract_id=contract_data.get("contract_id"),
                        acceptor=get_or_create_or_none(
                            "acceptor_id", contract_data, EveEntity
                        ),
                        assignee=get_or_create_or_none(
                            "assignee_id", contract_data, EveEntity
                        ),
                        availability=self.model.ESI_AVAILABILITY_MAP[
                            contract_data.get("availability")
                        ],
                        buyout=contract_data.get("buyout"),
                        collateral=contract_data.get("collateral"),
                        contract_type=self.model.ESI_TYPE_MAP.get(
                            contract_data.get("type"),
                            self.model.TYPE_UNKNOWN,
                        ),
                        date_accepted=contract_data.get("date_accepted"),
                        date_completed=contract_data.get("date_completed"),
                        date_expired=contract_data.get("date_expired"),
                        date_issued=contract_data.get("date_issued"),
                        days_to_complete=contract_data.get("days_to_complete"),
                        end_location=get_or_none(
                            "end_location_id", contract_data, Location
                        ),
                        for_corporation=contract_data.get("for_corporation"),
                        issuer_corporation=get_or_create_or_none(
                            "issuer_corporation_id", contract_data, EveEntity
                        ),
                        issuer=get_or_create_or_none(
                            "issuer_id", contract_data, EveEntity
                        ),
                        price=contract_data.get("price"),
                        reward=contract_data.get("reward"),
                        start_location=get_or_none(
                            "start_location_id", contract_data, Location
                        ),
                        status=self.model.ESI_STATUS_MAP[contract_data.get("status")],
                        title=contract_data.get("title", ""),
                        volume=contract_data.get("volume"),
                    )
                )

        self.bulk_create(new_contracts, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        return eve_entity_ids_from_objs(new_contracts)

    def _update_existing_contracts(
        self, character: Character, contracts_list: dict, contract_ids: set
    ) -> None:
        logger.info("%s: Updating %s contracts", character, len(contract_ids))
        update_contract_pks = list(
            self.filter(character=character, contract_id__in=contract_ids).values_list(
                "pk", flat=True
            )
        )
        contracts = self.in_bulk(update_contract_pks)
        for contract in contracts.values():
            contract_data = contracts_list.get(contract.contract_id)
            if contract_data:
                contract.acceptor = get_or_create_or_none(
                    "acceptor_id", contract_data, EveEntity
                )
                contract.acceptor_corporation = get_or_create_or_none(
                    "acceptor_corporation_id", contract_data, EveEntity
                )
                contract.date_accepted = contract_data.get("date_accepted")
                contract.date_completed = contract_data.get("date_completed")
                contract.status = self.model.ESI_STATUS_MAP[contract_data.get("status")]

        self.bulk_update(
            contracts.values(),
            fields=[
                "acceptor",
                "acceptor_corporation",
                "date_accepted",
                "date_completed",
                "status",
            ],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )


class CharacterContractBidManager(models.Manager):
    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def update_or_create_esi(self, character: Character, token: Token, contract):
        """Update or create contract bids for a contract from ESI."""
        if contract.contract_type != contract.TYPE_AUCTION:
            logger.warning(
                "%s, %s: Can not update bids. Wrong contract type.",
                character,
                contract.contract_id,
            )
            return

        logger.info(
            "%s, %s: Fetching contract bids from ESI", character, contract.contract_id
        )
        bids_data = (
            esi.client.Contracts.get_characters_character_id_contracts_contract_id_bids(
                character_id=character.eve_character.character_id,
                contract_id=contract.contract_id,
                token=token.valid_access_token(),
            ).results()
        )
        bids_list = {int(obj["bid_id"]): obj for obj in bids_data if "bid_id" in obj}
        self._update_or_create_objs(contract, bids_list)

    @transaction.atomic()
    def _update_or_create_objs(self, contract, bids_list: dict) -> Set[int]:
        """Update or create new contract objects from provided data.

        Return EveEntity IDs in newly created contracts.
        """
        incoming_ids = set(bids_list.keys())
        existing_ids = set(
            self.filter(contract=contract).values_list("bid_id", flat=True)
        )
        create_ids = incoming_ids.difference(existing_ids)
        if not create_ids:
            logger.info(
                "%s, %s: No new contract bids to add",
                contract.character,
                contract.contract_id,
            )
            return set()

        logger.info(
            "%s, %s: Storing %s new contract bids",
            contract.character,
            contract.contract_id,
            len(create_ids),
        )
        bids = [
            self.model(
                contract=contract,
                bid_id=bid.get("bid_id"),
                amount=bid.get("amount"),
                bidder=get_or_create_or_none("bidder_id", bid, EveEntity),
                date_bid=bid.get("date_bid"),
            )
            for bid_id, bid in bids_list.items()
            if bid_id in create_ids
        ]
        self.bulk_create(bids, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        new_entity_ids = {obj.bidder_id for obj in bids}
        return new_entity_ids


class CharacterContractItemQuerySet(models.QuerySet):
    def annotate_pricing(self) -> models.QuerySet:
        """Return qs with annotated price and total columns."""
        return (
            self.select_related("eve_type__market_price")
            .annotate(
                price=Case(
                    When(
                        raw_quantity=-2,
                        then=Value(None),
                    ),
                    default=F("eve_type__market_price__average_price"),
                )
            )
            .annotate(
                total=Case(
                    When(
                        raw_quantity=-2,
                        then=Value(None),
                    ),
                    default=ExpressionWrapper(
                        F("eve_type__market_price__average_price") * F("quantity"),
                        output_field=models.FloatField(),
                    ),
                )
            )
        )


class CharacterContractItemManagerBase(models.Manager):
    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def update_or_create_esi(self, character: Character, token: Token, contract):
        """Update or create contract items for a contract from ESI."""
        if contract.contract_type not in [
            contract.TYPE_ITEM_EXCHANGE,
            contract.TYPE_AUCTION,
        ]:
            logger.warning(
                "%s, %s: Can not update items. Wrong contract type.",
                character,
                contract.contract_id,
            )
            return

        items_data = self._fetch_data_from_esi(character, token, contract)
        self._update_or_create_objs(contract, items_data)

    def _fetch_data_from_esi(self, character: Character, token: Token, contract):
        logger.info(
            "%s, %s: Fetching contract items from ESI", character, contract.contract_id
        )
        my_esi = esi.client.Contracts
        items_data = my_esi.get_characters_character_id_contracts_contract_id_items(
            character_id=character.eve_character.character_id,
            contract_id=contract.contract_id,
            token=token.valid_access_token(),
        ).results()

        return items_data

    def _update_or_create_objs(self, contract, items_data):
        logger.info(
            "%s, %s: Storing %s contract items",
            self,
            contract.contract_id,
            len(items_data),
        )
        items = [
            self.model(
                contract=contract,
                record_id=item.get("record_id"),
                is_included=item.get("is_included"),
                is_singleton=item.get("is_singleton"),
                quantity=item.get("quantity"),
                raw_quantity=item.get("raw_quantity"),
                eve_type=get_or_create_esi_or_none("type_id", item, EveType),
            )
            for item in items_data
            if "record_id" in item
        ]
        # delete & create is fine here, since contracts items are never updated
        with transaction.atomic():
            self.filter(contract=contract).delete()
            self.bulk_create(items, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)


CharacterContractItemManager = CharacterContractItemManagerBase.from_queryset(
    CharacterContractItemQuerySet
)



from io import StringIO
from typing import Set
from unittest.mock import patch

from django.core.management import call_command
from django.db import IntegrityError
from django.test import TestCase
from eveuniverse.models import EveEntity

from app_utils.testing import create_fake_user

from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_contact,
)

MODULE_PATH = "memberaudit.management.commands.memberaudit_fix_eve_entities"


class TestDeleteUnresolved(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        user = create_fake_user(1001, "Bruce Wayne")
        cls.character = create_character(user.profile.main_character)

    def test_should_remove_unresolved_and_orphaned_only(self):
        # given
        EveEntity.objects.create(id=1)
        EveEntity.objects.create(
            id=1001, name="Bruce Wayne", category=EveEntity.CATEGORY_CHARACTER
        )
        contact = EveEntity.objects.create(id=2)
        create_character_contact(self.character, eve_entity=contact)

        # when
        call_command("memberaudit_fix_eve_entities", "--no-input", stdout=StringIO())

        # then
        expected = _fetch_eve_entity_ids()
        self.assertSetEqual(expected, {2, 1001})

    def test_should_skip_over_database_errors(self):
        # given
        EveEntity.objects.create(id=1)

        # when
        with patch(MODULE_PATH + ".EveEntity.delete") as mock:
            mock.side_effect = IntegrityError
            call_command(
                "memberaudit_fix_eve_entities", "--no-input", stdout=StringIO()
            )

        # then
        expected = _fetch_eve_entity_ids()
        self.assertSetEqual(expected, {1})

    def test_should_not_skip_over_other_errors(self):
        # given
        EveEntity.objects.create(id=1)

        # when
        with self.assertRaises(RuntimeError):
            with patch(MODULE_PATH + ".EveEntity.delete") as mock:
                mock.side_effect = RuntimeError
                call_command(
                    "memberaudit_fix_eve_entities", "--no-input", stdout=StringIO()
                )

        expected = _fetch_eve_entity_ids()
        self.assertSetEqual(expected, {1})


def _fetch_eve_entity_ids() -> Set[int]:
    return set(EveEntity.objects.values_list("id", flat=True))



"""Settings for Member Audit."""

from django.utils.translation import gettext_lazy as _

from app_utils.app_settings import clean_setting

MEMBERAUDIT_APP_NAME = clean_setting(
    "MEMBERAUDIT_APP_NAME", _("Member Audit"), required_type=str
)
"""Name of this app as shown in the Auth sidebar and page titles."""

MEMBERAUDIT_BULK_METHODS_BATCH_SIZE = clean_setting(
    "MEMBERAUDIT_BULK_METHODS_BATCH_SIZE", 500
)
"""Technical parameter defining the maximum number of objects processed per run
of Django batch methods, e.g. bulk_create and bulk_update.
"""

MEMBERAUDIT_DATA_RETENTION_LIMIT = clean_setting(
    "MEMBERAUDIT_DATA_RETENTION_LIMIT", default_value=360, min_value=7
)
"""Maximum number of days to keep historical data for mails, contracts and wallets.
Minimum is 7 day.
"""

MEMBERAUDIT_FEATURE_ROLES_ENABLED = clean_setting(
    "MEMBERAUDIT_FEATURE_ROLES_ENABLED", False
)
"""Feature flag to enable or disable the corporation roles feature."""

MEMBERAUDIT_DATA_EXPORT_MIN_UPDATE_AGE = clean_setting(
    "MEMBERAUDIT_DATA_EXPORT_MIN_UPDATE_AGE", 60
)
"""Minimum age of existing export file before next update can be started in minutes."""

MEMBERAUDIT_LOCATION_STALE_HOURS = clean_setting("MEMBERAUDIT_LOCATION_STALE_HOURS", 24)
"""Hours after a existing location (e.g. structure) becomes stale and gets updated
e.g. for name changes of structures.
"""

MEMBERAUDIT_MAX_MAILS = clean_setting("MEMBERAUDIT_MAX_MAILS", 250)
"""Maximum amount of mails fetched from ESI for each character."""

MEMBERAUDIT_NOTIFY_TOKEN_ERRORS = clean_setting("MEMBERAUDIT_NOTIFY_TOKEN_ERRORS", True)
"""When enabled will automatically notify users when their character has a token error.
But only once per character until the character is re-registered or this notification
is reset manually by admins.
"""

MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT = clean_setting(
    "MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT", 240
)
"""Default time in minutes after the last successful update at which a section
is considered stale and therefore needs to be updated.
All sections, which do not have a specific default value
and are not configured differently will use this value.

Tip: Please run the command ``memberaudit_stats`` to see the
currently effective configuration.
"""

MEMBERAUDIT_SECTION_STALE_MINUTES_CONFIG = clean_setting(
    "MEMBERAUDIT_SECTION_STALE_MINUTES_CONFIG", {}
)
"""Custom configuration of stale minutes for each section,
which will override the respective defaults.

Tip: Please run the command ``memberaudit_stats`` to see the
currently effective configuration.
"""

MEMBERAUDIT_SECTION_STALE_MINUTES_SECTION_DEFAULTS = {
    # former ring 1
    "location": 60,
    "online_status": 60,
    "ship": 60,
    "skill_queue": 60,
    # former ring 3
    "assets": 480,
    "attributes": 480,
    "corporation_history": 480,
    "fw_stats": 480,
    "loyalty": 480,
    "titles": 480,
}
"""Default values for stale minutes of specific sections.

:meta private:
"""

MEMBERAUDIT_STORE_ESI_DATA_ENABLED = clean_setting(
    "MEMBERAUDIT_STORE_ESI_DATA_ENABLED", False
)
"""Set to true to store incoming data from the ESI API to disk for debugging.

The data will be stored in JSON files under: `~/myauth/temp/memberaudit_log`.

Warning: Storing this data can quickly occupy a lot of disk space.
We strongly recommend to also define filters for sections and/or characters
to limit what data is stored.
"""

MEMBERAUDIT_STORE_ESI_DATA_SECTIONS = clean_setting(
    "MEMBERAUDIT_STORE_ESI_DATA_SECTIONS", []
)
"""List sections to filter storing debug data for. An empty list means all sections."""

MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS = clean_setting(
    "MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS", []
)
"""List character IDs to filter storing debug data for.
An empty list means all characters.
"""

MEMBERAUDIT_TASKS_HIGH_PRIORITY = clean_setting(
    "MEMBERAUDIT_TASKS_HIGH_PRIORITY", default_value=3, min_value=1, max_value=9
)
"""Priority for high priority tasks, e.g. user requests an action."""

MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS = clean_setting(
    "MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS", 2500
)
"""Technical parameter defining the maximum number of asset items processed in each pass
when updating character assets.
A higher value reduces duration, but also increases task queue congestion.
"""

MEMBERAUDIT_TASKS_LOW_PRIORITY = clean_setting(
    "MEMBERAUDIT_TASKS_LOW_PRIORITY", default_value=7, min_value=1, max_value=9
)
"""Priority for low priority tasks, e.g. updating characters."""

MEMBERAUDIT_TASKS_NORMAL_PRIORITY = clean_setting(
    "MEMBERAUDIT_TASKS_NORMAL_PRIORITY", default_value=5, min_value=1, max_value=9
)
"""Priority for normal tasks, e.g. updating characters."""

MEMBERAUDIT_TASKS_TIME_LIMIT = clean_setting("MEMBERAUDIT_TASKS_TIME_LIMIT", 7200)
"""Global timeout for tasks in seconds to reduce task accumulation during outages."""


####################
# Internal settings

# Timeout for caching objects when running tasks in seconds
MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT = clean_setting(
    "MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT", 600
)



# Generated by Django 4.0.7 on 2022-10-21 17:13

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0002_add_mining_ledger"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app, register, and view own characters",
                    ),
                    ("share_characters", "Can share his/her characters"),
                    ("finder_access", "Can access character finder feature"),
                    ("reports_access", "Can access reports feature"),
                    ("characters_access", "Can view characters owned by others"),
                    ("exports_access", "Can access data exports"),
                    ("view_shared_characters", "Can view shared characters"),
                    ("view_same_corporation", "Can view corporation characters"),
                    ("view_same_alliance", "Can view alliance characters"),
                    ("view_everything", "Can view all characters"),
                    (
                        "notified_on_character_removal",
                        "Notified when member drops character",
                    ),
                ),
            },
        ),
    ]



# flake8: noqa
"""This is a standalone scripts that generates many test characters for Member Audit."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
print("AA starting up...")
django.setup()

"""MAIN"""

import random
from typing import Any, Tuple

import requests
from tqdm import tqdm

from django.db.models import Max
from eveuniverse.models import EveEntity, EveType

from allianceauth.authentication.models import State
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

from memberaudit.models import Character, Location, SkillSet
from memberaudit.tests.testdata.constants import EveTypeId
from memberaudit.tests.testdata.factories import (
    create_character_asset,
    create_character_jump_clone,
    create_character_jump_clone_implant,
    create_character_skill_set_check,
    create_location,
)
from memberaudit.tests.utils import create_memberaudit_character

# configuration
CHARACTER_COUNT = 100  # max number of characters to generate
CORRUPTED_ASSETS_PER_CHARACTER = 100  # set to 0 to disable
CREATE_SKILL_SET_SKILLS = False

CORPORATION_IDS = [
    98615046,  # KarmaFleet University
    98627389,  # Alpha Academic
    98614492,  # Pandemic Horde High Sec
    98609240,  # Caladrius Hive
]  # eve characters are chosen randomly from these corporations

# setup globals
item_eve_type, _ = EveType.objects.get_or_create_esi(id=EveTypeId.VELDSPAR)
implant_eve_type, _ = EveType.objects.get_or_create_esi(
    id=EveTypeId.HIGH_GRADE_SNAKE_ALPHA
)

locations = Location.objects.filter(
    eve_type__isnull=False, eve_solar_system__isnull=False
)
valid_location_1 = locations.first()
valid_location_2 = locations.last()
if not valid_location_1 or valid_location_1 == valid_location_2:
    raise RuntimeError("Need two valid locations to exist")

character_entity = EveEntity.objects.filter(
    category=EveEntity.CATEGORY_CHARACTER
).first()
if not character_entity:
    raise RuntimeError("Need a character in eve entities to exist")

corporation_entity = EveEntity.objects.filter(
    category=EveEntity.CATEGORY_CORPORATION
).first()
if not corporation_entity:
    raise RuntimeError("Need a corporation in eve entities to exist")

next_location_id = (
    max(1_700_000_000_000, Location.objects.aggregate(Max("id"))["id__max"]) + 1
)


def main():
    my_state = _get_or_create_state_for_test_users()
    _delete_previous_test_characters(my_state)

    created_count = 0
    character_ids = _fetching_random_character_ids()[:CHARACTER_COUNT]
    for character_id in tqdm(
        character_ids, desc="Creating test characters", unit="objects"
    ):
        try:
            eve_character, created = get_or_create_eve_character(character_id)
        except OSError:
            continue
        if created:
            created_count += 1
            my_state.member_characters.add(eve_character)
            character = create_memberaudit_character(character_id, is_disabled=True)
            if CREATE_SKILL_SET_SKILLS:
                set_character_skill_set_checks(character)
            if CORRUPTED_ASSETS_PER_CHARACTER:
                add_invalid_assets_to_character(character)


def _delete_previous_test_characters(my_state):
    num, _ = EveCharacter.objects.filter(
        character_ownership__user__profile__state=my_state
    ).delete()
    if num > 0:
        print(f"Deleted stale test characters.")


def _get_or_create_state_for_test_users():
    my_state, created = State.objects.get_or_create(
        name="Test users", defaults={"priority": 75}
    )
    if created:
        basic_perm = AuthUtils.get_permission_by_name("memberaudit.basic_access")
        my_state.permissions.add(basic_perm)
    return my_state


def _fetching_random_character_ids():
    print(f"Selecting random character IDs from {len(CORPORATION_IDS)} corporations")
    character_ids = set()
    for corporation_id in CORPORATION_IDS:
        r = requests.get(f"https://evewho.com/api/corplist/{corporation_id}")
        if r.ok:
            data = r.json()
            character_ids |= {obj["character_id"] for obj in data["characters"]}

    character_ids = list(character_ids)
    random.shuffle(character_ids)
    return character_ids


def set_character_skill_set_checks(character: Character):
    for skill_set in SkillSet.objects.all():
        obj = create_character_skill_set_check(character=character, skill_set=skill_set)
        if random.choice([True, False, False]):
            skill = skill_set.skills.first()
            obj.failed_recommended_skills.add(skill)


def get_or_create_eve_character(character_id: int) -> Tuple[Any, bool]:
    """Get or create EveCharacter object."""
    try:
        return EveCharacter.objects.get(character_id=character_id), False
    except EveCharacter.DoesNotExist:
        return EveCharacter.objects.create_character(character_id=character_id), True


def add_invalid_assets_to_character(character: Character):
    global next_location_id

    for _ in range(CORRUPTED_ASSETS_PER_CHARACTER):
        next_location_id += 1
        asset = create_character_asset(
            character=character,
            eve_type=item_eve_type,
            location=valid_location_1,
            item_id=next_location_id,
        )
        invalid_location = create_location(id=asset.item_id)
        create_character_asset(
            character=character,
            eve_type=item_eve_type,
            location=invalid_location,
        )
        jump_clone = create_character_jump_clone(
            character=character, location=invalid_location
        )
        create_character_jump_clone_implant(
            jump_clone=jump_clone, eve_type=implant_eve_type
        )


if __name__ == "__main__":
    main()



import datetime as dt
from unittest.mock import patch

from django.contrib.auth.models import User
from django.test import RequestFactory
from django.urls import reverse
from django.utils.timezone import now

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import NoSocketsTestCase

from memberaudit.models import SkillSet
from memberaudit.tests.testdata.factories import (
    create_fitting_text,
    create_skill_set,
    create_skill_set_group,
)
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.views import admin

VIEWS_PATH = "memberaudit.views.admin"
MANAGERS_PATH = "memberaudit.managers.general"


@patch(VIEWS_PATH + ".messages", spec=True)
@patch(VIEWS_PATH + ".tasks", spec=True)
class TestCreateSkillSetFromFitting(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.superuser = User.objects.create_superuser("Superman")
        cls.fitting_text = create_fitting_text("fitting_tristan.txt")

    def test_should_open_page(self, mock_tasks, mock_messages):
        # given
        request = self.factory.get(
            reverse("memberaudit:admin_create_skillset_from_fitting")
        )
        request.user = self.superuser
        # when
        response = admin.admin_create_skillset_from_fitting(request)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_create_new_skillset(self, mock_tasks, mock_messages):
        # given
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_fitting"),
            data={"fitting_text": self.fitting_text},
        )
        request.user = self.superuser
        my_now = now()
        # when
        with patch(MANAGERS_PATH + ".now", lambda: my_now):
            response = admin.admin_create_skillset_from_fitting(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        self.assertTrue(mock_messages.info.called)
        self.assertEqual(SkillSet.objects.count(), 1)
        obj: SkillSet = SkillSet.objects.first()
        self.assertEqual(obj.last_modified_at, my_now)
        self.assertEqual(obj.last_modified_by, self.superuser)

    def test_should_overwrite_existing_skillset(self, mock_tasks, mock_messages):
        # given
        other_user = AuthUtils.create_user("nobody")
        old_dt = now() - dt.timedelta(days=10)
        skill_set = create_skill_set(
            name="Tristan - Standard Kite (cap stable)",
            last_modified_by=other_user,
            last_modified_at=old_dt,
        )
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_fitting"),
            data={"fitting_text": self.fitting_text, "can_overwrite": True},
        )
        request.user = self.superuser
        my_now = now()
        # when
        with patch(MANAGERS_PATH + ".now", lambda: my_now):
            response = admin.admin_create_skillset_from_fitting(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        self.assertTrue(mock_messages.info.called)
        skill_set.refresh_from_db()
        self.assertGreater(skill_set.skills.count(), 0)
        self.assertEqual(skill_set.last_modified_at, my_now)
        self.assertEqual(skill_set.last_modified_by, self.superuser)

    def test_should_create_new_skillset_and_assign_group(
        self, mock_tasks, mock_messages
    ):
        # given
        skill_set_group = create_skill_set_group()
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_fitting"),
            data={
                "fitting_text": self.fitting_text,
                "skill_set_group": skill_set_group.id,
            },
        )
        request.user = self.superuser
        # when
        response = admin.admin_create_skillset_from_fitting(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        skill_set = SkillSet.objects.first()
        self.assertIn(skill_set, skill_set_group.skill_sets.all())

    def test_should_create_new_skillset_with_custom_name(
        self, mock_tasks, mock_messages
    ):
        # given
        skill_set = create_skill_set(name="Tristan - Standard Kite (cap stable)")
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_fitting"),
            data={"fitting_text": self.fitting_text, "skill_set_name": "My-Name"},
        )
        request.user = self.superuser
        # when
        response = admin.admin_create_skillset_from_fitting(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        self.assertTrue(mock_messages.info.called)
        skill_set = SkillSet.objects.last()
        self.assertEqual(skill_set.name, "My-Name")


@patch(VIEWS_PATH + ".messages", spec=True)
@patch(VIEWS_PATH + ".tasks", spec=True)
class TestCreateSkillSetFromSkillPlan(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        cls.superuser = User.objects.create_superuser("Superman")
        cls.skill_plan_text = """
        Caldari Core Systems 5
        Caldari Strategic Cruiser 3
        """

    def test_should_open_page(self, mock_tasks, mock_messages):
        # given
        request = self.factory.get(
            reverse("memberaudit:admin_create_skillset_from_skill_plan")
        )
        request.user = self.superuser
        # when
        response = admin.admin_create_skillset_from_skill_plan(request)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_create_new_skillset(self, mock_tasks, mock_messages):
        # given
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_skill_plan"),
            data={
                "skill_set_name": "My Skill Set",
                "skill_plan_text": self.skill_plan_text,
            },
        )
        request.user = self.superuser
        my_now = now()
        # when
        with patch(MANAGERS_PATH + ".now", lambda: my_now):
            response = admin.admin_create_skillset_from_skill_plan(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        self.assertTrue(mock_messages.info.called)
        self.assertEqual(SkillSet.objects.count(), 1)
        obj: SkillSet = SkillSet.objects.first()
        self.assertEqual(obj.last_modified_at, my_now)
        self.assertEqual(obj.last_modified_by, self.superuser)

    def test_should_overwrite_existing_skillset_when_requested(
        self, mock_tasks, mock_messages
    ):
        # given
        other_user = AuthUtils.create_user("nobody")
        old_dt = now() - dt.timedelta(days=10)
        skill_set = create_skill_set(
            name="My Skill Set", last_modified_by=other_user, last_modified_at=old_dt
        )
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_skill_plan"),
            data={
                "skill_set_name": "My Skill Set",
                "skill_plan_text": self.skill_plan_text,
                "can_overwrite": True,
            },
        )
        request.user = self.superuser
        my_now = now()
        # when
        with patch(MANAGERS_PATH + ".now", lambda: my_now):
            response = admin.admin_create_skillset_from_skill_plan(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        self.assertTrue(mock_messages.info.called)
        skill_set.refresh_from_db()
        self.assertGreater(skill_set.skills.count(), 0)
        self.assertEqual(skill_set.last_modified_at, my_now)
        self.assertEqual(skill_set.last_modified_by, self.superuser)

    def test_should_create_new_skillset_and_assign_group(
        self, mock_tasks, mock_messages
    ):
        # given
        skill_set_group = create_skill_set_group()
        request = self.factory.post(
            reverse("memberaudit:admin_create_skillset_from_skill_plan"),
            data={
                "skill_set_name": "My Skill Set",
                "skill_plan_text": self.skill_plan_text,
                "skill_set_group": skill_set_group.id,
            },
        )
        request.user = self.superuser
        # when
        response = admin.admin_create_skillset_from_skill_plan(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_tasks.update_characters_skill_checks.delay.called)
        skill_set = SkillSet.objects.first()
        self.assertIn(skill_set, skill_set_group.skill_sets.all())



# Generated by Django 4.0.10 on 2023-04-19 20:03

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0006_add_localizations"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="character",
            options={
                "default_permissions": (),
                "verbose_name": "character",
                "verbose_name_plural": "characters",
            },
        ),
        migrations.AlterModelOptions(
            name="characterupdatestatus",
            options={
                "default_permissions": (),
                "verbose_name": "character update status",
                "verbose_name_plural": "character update status",
            },
        ),
        migrations.AlterModelOptions(
            name="compliancegroupdesignation",
            options={
                "default_permissions": (),
                "verbose_name": "compliance group designation",
                "verbose_name_plural": "compliance group designations",
            },
        ),
        migrations.AlterModelOptions(
            name="location",
            options={
                "default_permissions": (),
                "verbose_name": "location",
                "verbose_name_plural": "locations",
            },
        ),
        migrations.AlterModelOptions(
            name="skillset",
            options={
                "default_permissions": (),
                "verbose_name": "skill set",
                "verbose_name_plural": "skill sets",
            },
        ),
        migrations.AlterModelOptions(
            name="skillsetgroup",
            options={
                "default_permissions": (),
                "verbose_name": "skill set group",
                "verbose_name_plural": "skill set groups",
            },
        ),
        migrations.AlterModelOptions(
            name="skillsetskill",
            options={
                "default_permissions": (),
                "verbose_name": "skill",
                "verbose_name_plural": "skills",
            },
        ),
    ]
